#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

import inspect
import copy

def function():
    return inspect.currentframe().f_back.f_code.co_name

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class Item:
    def __init__(self, kind, price):
        self.kind = kind
        self.code = 0
        self.price = price
    def __repr__(self):
        return '«{}/{}/{}€»'.format(self.kind, self.code, self.price)

class Shop:
    def __init__(self, name):
        self.__name = name
        self.__next_code = 0
        self.__stock = {}
    #
    def name(self):
        return self.__name
    #
    def stock_dangerous(self):
        return self.__stock # !!! encapsulation breakage here !!!
    #
    def stock(self):
        return copy.deepcopy(self.__stock) # really expensive, for safety
    #
    def receive(self, item):
        self.__next_code += 1
        item.code = self.__next_code
        l = self.__stock.get(item.kind, [])
        l.append(item)
        self.__stock[item.kind] = l
    #
    def remove(self, item):
        items = self.__stock[item.kind]
        for (idx, it) in enumerate(items):
            if it.code == item.code:
                items.pop(idx)
                break
    #
    def show(self):
        txt='{!r}:'.format(self.__name)
        for (kind, items) in self.__stock.items():
            txt += '\n• {} {}'.format(kind, items)
        return txt

class Basket:
    def __init__(self):
        self.__items = []
    #
    def choose_item(self, shop, kind):
        self.__items.append((shop, shop.stock()[kind][-1]))
    #
    def buy(self):
        items = []
        while self.__items:
            (shop, item) = self.__items.pop()
            shop.remove(item)
            items.append(item)
        return items

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def test_encapsulation_breakage():
    print('~~~~~~~~ {}() ~~~~~~~~'.format(function()))
    shop = Shop('Everything for home')
    for i in range(2):
        shop.receive(Item('lamp', 25.4))
    for i in range(3):
        shop.receive(Item('plate', 4.25))
    for i in range(4):
        shop.receive(Item('knife', 7.6))
    print(shop.show())
    stock = shop.stock()
    stock['lamp'].append(stock['knife'].pop()) # no infuence on shop
    print(shop.show())
    stock = shop.stock_dangerous()
    stock['lamp'].append(stock['knife'].pop()) # !!! this shop is messy !!!
    print(shop.show())

def test_normal_usage():
    print('~~~~~~~~ {}() ~~~~~~~~'.format(function()))
    shop1 = Shop('Everything for home')
    for i in range(2):
        shop1.receive(Item('lamp', 25.4))
    for i in range(3):
        shop1.receive(Item('plate', 4.25))
    for i in range(4):
        shop1.receive(Item('knife', 7.6))
    shop2 = Shop('Look pretty')
    for i in range(3):
        shop2.receive(Item('scarf', 11.9))
    for i in range(2):
        shop2.receive(Item('socks', 3.25))
    print(shop1.show())
    print(shop2.show())
    basket1 = Basket()
    basket1.choose_item(shop1, 'knife')
    basket1.choose_item(shop1, 'lamp')
    basket1.choose_item(shop2, 'scarf')
    print('bought with basket1', basket1.buy())
    basket2 = Basket()
    basket2.choose_item(shop1, 'plate')
    basket2.choose_item(shop1, 'lamp')
    basket2.choose_item(shop2, 'socks')
    print('bought with basket2', basket2.buy())
    print(shop1.show())
    print(shop2.show())

def test_conflict():
    print('~~~~~~~~ {}() ~~~~~~~~'.format(function()))
    shop = Shop('Everything for home')
    for i in range(2):
        shop.receive(Item('lamp', 25.4))
    for i in range(3):
        shop.receive(Item('plate', 4.25))
    for i in range(4):
        shop.receive(Item('knife', 7.6))
    print(shop.show())
    basket1 = Basket()
    basket1.choose_item(shop, 'knife')
    basket1.choose_item(shop, 'lamp')
    basket2 = Basket()
    basket2.choose_item(shop, 'plate')
    basket2.choose_item(shop, 'lamp')
    print('bought with basket1', basket1.buy()) # !!! the same lamp ...
    print('bought with basket2', basket2.buy()) # ... is bought twice !!!
    print(shop.show())

if __name__=='__main__':
    test_encapsulation_breakage()
    test_normal_usage()
    test_conflict()

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
