use std::io::Write;

// exemple d'association avec non-parteageabilité
pub struct ExclusiveCountBytes<'a> {
    file: &'a mut std::fs::File,
    count: usize,
}

impl<'a> ExclusiveCountBytes<'a> {
    pub fn new(file: &'a mut std::fs::File) -> Self {
        Self { file, count: 0 }
    }

    pub fn write(
        &mut self,
        txt: &str,
    ) {
        self.file.write_all(txt.as_bytes()).unwrap();
        self.count += txt.len();
    }

    pub fn count(&self) -> usize {
        self.count
    }
}

// exemple d'association avec parteageabilité
pub struct SharedCountBytes<'a> {
    file: &'a std::cell::RefCell<std::fs::File>,
    count: usize,
}

impl<'a> SharedCountBytes<'a> {
    pub fn new(file: &'a std::cell::RefCell<std::fs::File>) -> Self {
        Self { file, count: 0 }
    }

    pub fn write(
        &mut self,
        txt: &str,
    ) {
        self.file.borrow_mut().write_all(txt.as_bytes()).unwrap();
        self.count += txt.len();
    }

    pub fn count(&self) -> usize {
        self.count
    }
}
