pub mod association;
pub mod composition;
pub mod shop;

use association::*;
use composition::*;
use shop::*;

// ~~~~~~~~~~~~~~ TESTS COMPOSITION ~~~~~~~~~~~~~~ //
fn test_composition_house() {
    println!("~~~~~~~~ test_composition_house() ~~~~~~~~");
    println!("Pour faire ce test il faut décommenter le code");
    // Ce test montre que Rust détecte une référence
    // qui ne référence rien du tout déjà à la compilation
    // pour l'essayer le décommenter, puis le recommenter

    /*
    fn house_problem() -> &Room {
        let mut hs = House::new();
        for i in 0..4 {
            hs.add("bedroom".to_owned(), 16 + i);
        }

        println!("{:?}", hs);

        let br = hs.get_bedroom().unwrap();

        println!("returning {} {}", br.kind, br.square_meters);
        br
    }

    let br = house_problem();

    println!("obtained {} {}", br.kind, br.square_meters);
    */
}

fn test_composition_1_01() {
    println!("~~~~~~~~ test_composition_1_01() ~~~~~~~~");
    let red = Color {
        red: 255,
        green: 0,
        blue: 0,
    };

    let pos = Point { x: 10.0, y: 9.0 };

    let rect1 = Rectangle::new(pos, red.clone(), None, 20.0, 30.0);

    let rect2 = Rectangle::new(
        Point { x: 1.0, y: 9.0 },
        red,
        Some(Color {
            red: 0,
            green: 255,
            blue: 0,
        }),
        20.0,
        30.0,
    );

    println!("Rectangle rect1");
    rect1.draw();
    println!("Rectangle rect2");
    rect2.draw();
}

fn test_composition_plus() {
    println!("~~~~~~~~ test_composition_plus() ~~~~~~~~");
    let mut plt = Palette::new(3);
    for i in 0..4 {
        plt.add(Color {
            red: i,
            green: i,
            blue: i,
        });
    }

    println!("Palette plt");
    plt.draw();

    let img = Image::new_black(2, 2);

    println!("Image img");
    img.draw();
}

// ~~~~~~~~~~~~~~ TESTS ASSOCIATION ~~~~~~~~~~~~~~ //
fn test_exclusive_count_bytes() {
    println!("~~~~~~~~ test_exclusive_count_bytes() ~~~~~~~~");
    let mut f = std::fs::File::create("output_rs_e.txt").unwrap();
    let mut c1 = ExclusiveCountBytes::new(&mut f);
    c1.write("ceci\n");
    c1.write("cela\n");
    println!("{}", c1.count());

    // let mut c2 = ExclusiveCountBytes::new(&mut f);
    // c2.write("autre chose\n");

    c1.write("encore\n");
    println!("{}", c1.count());
}

fn test_shared_count_bytes() {
    println!("~~~~~~~~ test_shared_count_bytes() ~~~~~~~~");
    let f = std::cell::RefCell::new(
        std::fs::File::create("output_rs_s.txt").unwrap(),
    );
    let mut c = SharedCountBytes::new(&f);
    c.write("ceci\n");
    c.write("cela\n");
    println!("{}", c.count());

    let mut c2 = SharedCountBytes::new(&f);
    c2.write("autre chose\n");

    // drop(f);

    c.write("encore\n");
    println!("{}", c.count());
}

// ~~~~~~~~~~~~~~ TESTS ASSOCIATION ~~~~~~~~~~~~~~ //
// pour les types déclarés dans shop.rs
fn test_normal_usage() -> Result<(), Box<dyn std::error::Error>> {
    println!("~~~~~~~~ test_normal_usage() ~~~~~~~~");
    let mut shop1 = Shop::new("Everything for home".to_owned());
    for _ in 0..2 {
        shop1.receive(Item::new("lamp".to_owned(), 25.4));
    }
    for _ in 0..3 {
        shop1.receive(Item::new("plate".to_owned(), 4.25));
    }
    for _ in 0..4 {
        shop1.receive(Item::new("knife".to_owned(), 7.60));
    }
    let mut shop2 = Shop::new("Look pretty".to_owned());
    for _ in 0..3 {
        shop2.receive(Item::new("scarf".to_owned(), 11.9));
    }
    for _ in 0..2 {
        shop2.receive(Item::new("socks".to_owned(), 3.25));
    }
    println!("{}", shop1.show());
    println!("{}", shop2.show());
    let mut basket1 = Cart::new();
    basket1.choose_item(&shop1, "knife")?;
    basket1.choose_item(&shop1, "lamp")?;
    basket1.choose_item(&shop2, "scarf")?;
    println!("bought with basket1 {:?}", basket1.buy());
    let mut basket2 = Cart::new();
    basket2.choose_item(&shop1, "plate")?;
    basket2.choose_item(&shop1, "lamp")?;
    basket2.choose_item(&shop2, "socks")?;
    println!("bought with basket2 {:?}", basket2.buy());
    println!("{}", shop1.show());
    println!("{}", shop2.show());
    Ok(())
}

fn test_conflict() -> Result<(), Box<dyn std::error::Error>> {
    println!("~~~~~~~~ test_conflict() ~~~~~~~~");
    let mut shop = Shop::new("Everything for home".to_owned());
    for _ in 0..2 {
        shop.receive(Item::new("lamp".to_owned(), 25.4));
    }
    for _ in 0..3 {
        shop.receive(Item::new("plate".to_owned(), 4.25));
    }
    for _ in 0..4 {
        shop.receive(Item::new("knife".to_owned(), 7.60));
    }
    println!("{}", shop.show());
    let mut basket1 = Cart::new();
    basket1.choose_item(&shop, "knife")?;
    basket1.choose_item(&shop, "lamp")?;
    let mut basket2 = Cart::new();
    basket2.choose_item(&shop, "plate")?;
    basket2.choose_item(&shop, "lamp")?;
    println!("bought with basket1 {:?}", basket1.buy());
    println!("bought with basket2 {:?}", basket2.buy()); // no lamp
    println!("{}", shop.show());
    Ok(())
}

fn main() {
    //~~~~~~~~ test détection incohérence ~~~~~~~~
    test_composition_house();
    //~~~~~~~~ test composition ~~~~~~~~
    test_composition_1_01();
    test_composition_plus();

    //~~~~~~~~ test association ~~~~~~~~
    test_exclusive_count_bytes();
    test_shared_count_bytes();

    //~~~~~~~~ test association shop ~~~~~~~~
    for fnct in [test_normal_usage, test_conflict] {
        if let Err(err) = fnct() {
            println!("{:?}", err);
        }
    }
}
